// openplanet doesnt like comments on same line as #if statement?
// just here for visual studio code intellisense help
#if __INTELLISENSE__
#include "STMCPTimes.h"
#include "Settings.as"
#include "cppIntellisense.h"
#endif

// timing
int prevLapRaceTime = 0;
// last waypoint cp id
int lastCP = 0;
int currentLap = 0;
// incrementing cp
int currCP = 0;

int finishRaceTime = 0;
int cpCycleFromTime = 0;

// map data
int numCps = 0;
string currentMap;
int numLaps = 0;

class Record {
  int checkpointId = -1;
  int time;

  Record(int checkpointId, int time) {
    this.checkpointId = checkpointId;
    this.time = time;
  }

  Record @opAssign(const Record &in record) {
    checkpointId = record.checkpointId;
    time = record.time;
    return this;
  }
}

// storage
// lap times
Record @[] bestTimesRec;


// extra
bool waitForCarReset = true;
bool hasPlayerRaced = false;
bool resetData = true;
int playerStartTime = -1;
bool firstLoad = false;
bool isMultiLap = false;
bool isFinished = false;

// font
string loadedFontFace = "";
int loadedFontSize = 0;
UI::Font @font = null;

// turbo extra's
bool hasFinishedMap = false;

void debugText(const string &in text) {
    // print(text);
}

void Main() {
  LoadFont();
  // waitForCarReset = GetCurrentCheckpoint()!= 0;
  playerStartTime = GetPlayerStartTime();
  // currentMap = GetMapId();
}



// clears data, waits till a reset to start happens
void ResetCommon() {
  waitForCarReset = true;
  ResetRace();
  bestTimesRec = {};

  hasFinishedMap = false;

}

void ResetRace() {
  bestTimesRec = {};
  isFinished = false;
  currentLap = 0;
  currCP = 0;
  finishRaceTime = 0;
  prevLapRaceTime = 0;
  cpCycleFromTime = Time::get_Now();

#if TMNEXT
  lastCP = GetSpawnCheckpoint();
#endif
}

void Update(float dt) {
  // have we changed map?
  if (currentMap != GetMapId()) {
    debugText("map mismatch");
    debugText("MapID: " + GetMapId());
    debugText("CurrentMap: " + currentMap);

    // we have left the map, lets save our data before we lose it
    if (currentMap != "") {
    }

    // reset file names
    currentMap = "";

    // reset important variables
    ResetCommon();

    // check for a valid map
    if (currentMap == "" || GetMapId() == "") {
      debugText("map found - " + GetMapName());
      // waitForCarReset = false;
      playerStartTime = GetPlayerStartTime();

      currentMap = GetMapId();


      UpdateWaypoints();
    }

  }
  // wait for the car to be back at starting checkpoint
  if (waitForCarReset) {
#if TMNEXT
    // when waiting for a reset, Player race time ends up at a - value for the
    // countdown before starting the lap
    waitForCarReset = GetCurrentPlayerRaceTime() >= 0;
#endif
    return;
  }

  // wait for car to be driveable to do our final reset
  if (resetData) {
    if (IsPlayerReady()) {
      debugText("Running reset");
      resetData = false;

      ResetRace();

      playerStartTime = GetActualPlayerStartTime();

      if (numCps <= 1) {
        UpdateWaypoints();
      }

      hasPlayerRaced = true;

      debugText("Ready to read checkpoints");
    }
    return;
  } else {
    // if our car no longer became valid? reset?
    if (!IsPlayerReady()) {
      debugText("Car no longer valid..");
      resetData = true;

      return;

    }
  }

  int cp = GetCurrentCheckpoint();


  // have we changed checkpoint?
  if (cp != lastCP) {
    
    debugText("- Checkpoint change " + lastCP + " to " + cp);

    lastCP = cp;

    int raceTime = GetPlayerCheckpointTime();

    // debugText(multiLapTime);
    int deltaTime = raceTime - prevLapRaceTime;

    debugText("Delta time: " + deltaTime);
    debugText("Race time: " + raceTime);


    if (raceTime <= 0 || deltaTime <= 0) {
      debugText("Checkpoint time negative..");
      // not sure if this is necessary anymore?
#if TMNEXT
      waitForCarReset = true;
#endif
      return;
    }

    // add our time (best current time for run)
    if (isMultiLap && IsWaypointFinish(cp)) {
      prevLapRaceTime = raceTime;
      CreateOrUpdateBestTime(cp, deltaTime);
      currCP++;
    }
    

    // check for finish
#if TMNEXT
    if (IsWaypointFinish(cp)) {
#endif
      hasFinishedMap = true;
      currentLap++;
      if (isMultiLap) {
        debugText("Lap finish: " + currentLap + "/" + numLaps);
      }

#if TMNEXT
      if (!isMultiLap || currentLap == numLaps) {
#endif
        debugText("Race Finished");

        waitForCarReset = true;
        resetData = true;
        isFinished = true;
        finishRaceTime = raceTime;
        cpCycleFromTime = Time::get_Now();
        
      } // race finish
    }   // lap finish


  } // change cp
}


void CreateOrUpdateBestTime(int checkpoint, int time) {
  int recIndex = -1;
    Record rec(checkpoint, time);
    bestTimesRec.InsertLast(rec);
  
}



#if TMNEXT
CSmArenaClient @GetPlayground() {
  CSmArenaClient @playground = cast<CSmArenaClient>(GetApp().CurrentPlayground);
  return playground;
}
#endif

CSmArenaRulesMode @GetPlaygroundScript() {
  CSmArenaRulesMode @playground =
      cast<CSmArenaRulesMode>(GetApp().PlaygroundScript);
  return playground;
}

int GetCurrentGameTime() {
  auto playground = GetPlayground();
  if (playground is null || playground.Interface is null ||
      playground.Interface.ManialinkScriptHandler is null) {
    return -1;
  }
  // CSmArenaInterfaceManialinkScripHandler@ aimsh =
  // cast<CSmArenaInterfaceManialinkScripHandler>(playground.Interface.ManialinkScriptHandler
  // ); return aimsh.ArenaNow;
  return playground.Interface.ManialinkScriptHandler.GameTime;
}

#if TMNEXT
CSmPlayer @GetPlayer() {
  auto playground = GetPlayground();
  if (playground is null || playground.GameTerminals.Length != 1) {
    return null;
  }

  return cast<CSmPlayer>(playground.GameTerminals[0].GUIPlayer);
}
#endif


CSmScriptPlayer @GetPlayerScript() {
  CSmPlayer @smPlayer = GetPlayer();
  if (smPlayer is null) {
    return null;
  }
  return cast<CSmScriptPlayer>(smPlayer.ScriptAPI);
}

bool IsPlayerReady() {
  CSmScriptPlayer @smPlayerScript = GetPlayerScript();
  if (smPlayerScript is null) {
    return false;
  }
  return GetCurrentPlayerRaceTime() >= 0 &&
         smPlayerScript.Post == CSmScriptPlayer::EPost::CarDriver &&
         GetSpawnCheckpoint() != -1;
}


// current time for race
// not accurate for ui
int GetCurrentPlayerRaceTime() {
  return GetCurrentGameTime() - GetPlayerStartTime();
}

// chooses between GetUICheckpointTime or GetCurrentPlayerRaceTime
// called directly after a checkpoint changed
int GetPlayerCheckpointTime() {
#if TMNEXT
  int raceTime;
  int estRaceTime = GetCurrentPlayerRaceTime();
  int uiRaceTime = GetUICheckpointTime();
  if (uiRaceTime == -1) {
    raceTime = estRaceTime;
  } else {
    raceTime = uiRaceTime;
  }
  // debugText("ui: " + Time::Format(uiRaceTime) + " - norm: " +
  // Time::Format(raceTime));
  return raceTime;
#endif
}


#if TMNEXT
int GetPlayerStartTime() {
  CSmPlayer @smPlayer = GetPlayer();
  if (smPlayer is null) {
    return -1;
  }
  return smPlayer.StartTime;
}
#endif

int GetActualPlayerStartTime() {
  return GetPlayerStartTime() - GetCurrentPlayerRaceTime();
}

#if TMNEXT
int GetUICheckpointTime() {
  CGameCtnNetwork @network = GetApp().Network;
  if (network is null) {
    return -1;
  }
  CGameManiaAppPlayground @appPlayground = network.ClientManiaAppPlayground;
  if (appPlayground is null) {
    return -1;
  }

  // search? instead of hardcode?
  CGameUILayer @raceUILayer = appPlayground.UILayers[8];
  if (raceUILayer is null) {
    return -1;
  }
  CGameManialinkPage @linkPage = raceUILayer.LocalPage;
  if (linkPage is null) {
    return -1;
  }
  CGameManialinkControl @race_Checkpoint = linkPage.GetClassChildren_Result[0];
  if (race_Checkpoint is null ||
      race_Checkpoint.ControlId != "Race_Checkpoint") { // validation
    return -1;
  }
  CGameManialinkControl @frame_checkpoint =
      cast<CGameManialinkFrame>(race_Checkpoint).Controls[0];
  if (frame_checkpoint is null) {
    return -1;
  }
  CGameManialinkControl @frame_race =
      cast<CGameManialinkFrame>(frame_checkpoint).Controls[0];
  if (frame_race is null) {
    return -1;
  }
  CGameManialinkControl @frame_race_time =
      cast<CGameManialinkFrame>(frame_race).Controls[0];
  if (frame_race_time is null ||
      cast<CGameManialinkFrame>(frame_race_time).Controls.Length != 2) {
    return -1;
  }
  CGameManialinkControl @label_race_time_ctrl =
      cast<CGameManialinkFrame>(frame_race_time).Controls[1];
  if (label_race_time_ctrl is null) {
    return -1;
  }
  CGameManialinkLabel @label_race_time =
      cast<CGameManialinkLabel>(label_race_time_ctrl);
  if (label_race_time is null) {
    return -1;
  }
  // debugText(label_race_time.Value);

  return ConvertStringToTime(label_race_time.Value);
}
#endif

int ConvertStringToTime(const string &in input) {
  string[] seconds = SplitString(input, ":");
  if (seconds.Length != 2) {
    return -1;
  }
  string[] microSecond = SplitString(seconds[1], ".");
  if (microSecond.Length != 2) {
    return -1;
  }
  int micro = Text::ParseInt(microSecond[1]);
  int second = Text::ParseInt(microSecond[0]);
  int minute = Text::ParseInt(seconds[0]);
  second *= 1000;
  minute *= 1000 * 60;
  int result = micro + second + minute;
  // debugText(Time::Format(1000));
  // debugText(Time::Format(result));
  return result;
}

// splits strings, removes split character
string[] SplitString(const string &in input, const string &in split) {
  // how to just pass a char??
  string current = "";
  string[] output;
  for (int i = 0; i < input.Length; i++) {
    if (input[i] == split[0]) {
      output.InsertLast(current);
      current = "";
    } else {
      // how to do just add a char?
      current += " ";
      current[current.Length - 1] = input[i];
    }
  }
  output.InsertLast(current);
  return output;
}

int GetCurrentCheckpoint() {
#if TMNEXT
  CSmPlayer @smPlayer = GetPlayer();
  if (smPlayer is null) {
    return -1;
  }
  return smPlayer.CurrentLaunchedRespawnLandmarkIndex;
#endif
}

string GetMapName() {
#if TMNEXT || MP4
  if (GetApp().RootMap is null) {
    return "";
  }
  return GetApp().RootMap.MapName;
#endif
}

string GetMapId() {
#if TMNEXT || MP4
  if (GetApp().RootMap is null) {
    return "";
  }
  return GetApp().RootMap.IdName;
#endif
}

#if TMNEXT
bool IsWaypointFinish(int index) {
  if (index == -1) {
    return false;
  }
  auto playground = GetPlayground();
  if (playground is null) {
    return false;
  }
  MwFastBuffer<CGameScriptMapLandmark @> landmarks =
      playground.Arena.MapLandmarks;
  if (index >= int(landmarks.Length)) {
    return false;
  }
  return landmarks[index].Waypoint !is null ? landmarks[index].Waypoint.IsFinish
                                            : false;
}

bool IsWaypointStart(int index) { return GetSpawnCheckpoint() == index; }

bool IsWaypointValid(int index) {
  if (index == -1) {
    return false;
  }
  auto playground = GetPlayground();
  MwFastBuffer<CGameScriptMapLandmark @> landmarks =
      playground.Arena.MapLandmarks;
  if (index >= int(landmarks.Length)) {
    return false;
  }
  bool valid = false;
  // valid |= (landmarks[index].Waypoint != null ?
  // landmarks[index].Waypoint.IsFinish : false); valid |= (landmarks[index].Tag
  // == "Checkpoint");
  if (landmarks[index].Waypoint !is null ? landmarks[index].Waypoint.IsFinish
                                         : false)
    valid = true;
  if ((landmarks[index].Tag == "Checkpoint"))
    valid = true;
  return valid;
}

int GetSpawnCheckpoint() {
  CSmPlayer @smPlayer = GetPlayer();
  if (smPlayer is null) {
    return -1;
  }
  return smPlayer.SpawnIndex;
}
#endif

void UpdateWaypoints() {

#if TMNEXT
  numCps = 1; // one for finish
#endif
  numLaps = 1;
  isMultiLap = false;

  array<int> links = {};
  bool strictMode = true;

#if TMNEXT
  auto playground = GetPlayground();
  if (playground is null || playground.Arena is null ||
      playground.Arena.Rules is null) {
    debugText("map not ready for waypoint read?");
    return;
  }

  auto map = playground.Map;
#endif
  if (map is null) {
    return;
  }
  numLaps = map.TMObjective_NbLaps;
  isMultiLap = map.TMObjective_IsLapRace;
  debugText("Map Laps: " + numLaps + " Is MultiLap: " + isMultiLap);
#if TMNEXT
  MwFastBuffer<CGameScriptMapLandmark @> landmarks =
      playground.Arena.MapLandmarks;
  for (uint i = 0; i < landmarks.Length; i++) {
    if (landmarks[i].Waypoint is null) {
      continue;
    }
    if (landmarks[i].Waypoint.IsMultiLap) {
      // isMultiLap = true;
      continue;
    }
    if (landmarks[i].Waypoint.IsFinish) {
      // numCps++;
      continue;
    }
    // if(landmarks[i].Tag == "Spawn"){
    // }
    //  we have a CP, but we don't know if it is Linked or not
    if (landmarks[i].Tag == "Checkpoint") {
      numCps++;
    } else if (landmarks[i].Tag == "LinkedCheckpoint") {
      if (links.Find(landmarks[i].Order) < 0) {
        numCps++;
        links.InsertLast(landmarks[i].Order);
      }
    } else {
      // this waypoint looks like a CP, acts like a CP, but is not called a CP.
      if (strictMode) {
        warn("The current map, " + string(playground.Map.MapName) + " (" +
             playground.Map.IdName +
             "), is not compliant with checkpoint naming rules.");
      }
      numCps++;
      strictMode = false;
    }
  }

  if (isMultiLap && numLaps == 1) {
    numCps -= 1;
    isMultiLap = false;
  }

  hasFinishedMap = true;
#endif
}



void OnSettingsChanged() {
  //
}

// modified https://github.com/Phlarx/tm-ultimate-medals
void Render() {
  auto app = cast<CTrackMania>(GetApp());

#if TMNEXT || MP4
  auto map = app.RootMap;
#endif




  int timeWidth = 53;
  int deltaWidth = 60;

  // number of cols we show lap data for
  int dataCols = 0;
  dataCols += 1; // Lap
  dataCols += 1; // Time


  if (!isMultiLap) {
    return;
  }

  if (hideWithIFace) {
    auto playground = app.CurrentPlayground;
    if (playground is null || playground.Interface is null ||
        !UI::IsGameUIVisible()) {
      return;
    }
  }


  bool invalidRun = !isFinished && waitForCarReset;

  if (invalidRun) {
    UI::PushStyleColor(UI::Col::Text, vec4(1.0, 0.0, 0.0, 1.0));
  }

  // if (isDisplayingSomething && windowVisible && map!is null &&
  // map.MapInfo.MapUid != "" && app.Editor is null) {
  if (windowVisible && map !is null &&
      map.MapInfo.MapUid != "") {
    if (lockPosition) {
      UI::SetNextWindowPos(int(anchor.x), int(anchor.y), UI::Cond::Always);
    } else {
      UI::SetNextWindowPos(int(anchor.x), int(anchor.y),
                           UI::Cond::FirstUseEver);
    }

    int windowFlags =
        UI::WindowFlags::NoTitleBar | UI::WindowFlags::NoCollapse |
        UI::WindowFlags::AlwaysAutoResize | UI::WindowFlags::NoDocking;
    if (!UI::IsOverlayShown()) {
      windowFlags |= UI::WindowFlags::NoInputs;
    }

    UI::PushFont(font);

    UI::Begin("LapTimes", windowFlags);

    if (!lockPosition) {
      anchor = UI::GetWindowPos();
    }

    UI::BeginGroup();


    if (UI::BeginTable("info", 1, UI::TableFlags::SizingFixedFit)) {

       if (showAverageLap) {
        UI::TableNextColumn();
        int sum = 0;
        for (uint i = 0; i < bestTimesRec.Length ; i++) {
          sum += bestTimesRec[i].time;
        }

        int averageLapTime = (bestTimesRec.Length > 0) ?  sum/bestTimesRec.Length : 0;
        string text = (averageLapTime > 0) ? Time::Format(averageLapTime) : '-';

        UI::Text("Average lap: " + text);
      }
      UI::EndTable();
    }

    if (dataCols != 0 &&
        UI::BeginTable("table", dataCols, UI::TableFlags::SizingFixedFit)) {

          UI::TableNextColumn();
          SetMinWidth(0);
          UI::Text("Lap");


          UI::TableNextColumn();
          SetMinWidth(timeWidth);
          UI::Text("Time");
         // end header

      uint minNum =
          Math::Min(bestTimesRec.Length, Math::Abs(numCheckpointsOnScreen));
      uint offset = 0;
      bool cycleThoughCheckpoints = isFinished || !hasPlayerRaced;

      if (minNum < bestTimesRec.Length) {
        offset = bestTimesRec.Length - minNum;

        if (cycleThoughCheckpoints) {
          int postOffset =
              int((Time::get_Now() - cpCycleFromTime) /
                  (Math::Max(0.001f, finishedCheckpointCycleSpeed) * 1000.0f));
          offset += Math::Max(0, postOffset);
        }
      }

      for (uint q = 0; q < minNum; q++) {
        uint i = offset + q;

        if (isMultiLap || cycleThoughCheckpoints) {
          i = i % bestTimesRec.Length;
        }
        UI::TableNextRow();

        // Lap
          UI::TableNextColumn();
          UI::Text("" + (i + 1));
          
        // Time
          UI::TableNextColumn();
          if (bestTimesRec.Length > i) {
            UI::Text(Time::Format(bestTimesRec[i].time));
          }

      } // end lap data

      UI::EndTable();
    }

    UI::EndGroup();

    UI::End();

    UI::PopFont();
  }

  if (invalidRun) {
    UI::PopStyleColor();
  }
}



void SetMinWidth(int width) {
  UI::PushStyleVar(UI::StyleVar::ItemSpacing, vec2(0, 0));
  UI::Dummy(vec2(width, 0));
  UI::PopStyleVar();
}

void LoadFont() {
  string fontFaceToLoad = fontFace.Length == 0 ? "DroidSans.ttf" : fontFace;
  if (fontFaceToLoad != loadedFontFace || fontSize != loadedFontSize) {
    @font = UI::LoadFont(fontFaceToLoad, fontSize);
    if (font !is null) {
      loadedFontFace = fontFaceToLoad;
      loadedFontSize = fontSize;
    }
  }
}

// HELPERS

int BoolToInt(bool value) {
  if (value) {
    return 1;
  }
  return 0;
}

// UIModule_Race_TimeGap
// Network
// ClientManiaAppPlayground
// UILayers
// 10
// LocalPage
// GetClassChidren_Result
// Frame 0
// Controls 0
// Frame 0
// Control
// childs 0
// childs 0
// childs 1
// childs 3 (bottom right time diff window 4th player down)

// UIModule_Race_Checkpoint
// Network
// ClientManiaAppPlayground
// UILayers
// 8
// LocalPage
// GetClassChidren_Result
// Frame 0
// Controls 0
// Frame 0
// Controls 0
// Frame 0
// Controls 0
// Frame 0
// Controls 0
// Frame 0
// Controls 0
// controls 1 (label)
